from typing import Callable, TypeAlias, Iterable, TypeVar, Protocol, Optional, ParamSpec
from requests import get
from osmium_archive.algebra.types import ArchiveLink, find_links


class TagType(Protocol):
    """
    Protocol used for all analysers. A tag represents a html element within a html document.
    """
    name: str
    """
    Name of the tag.
    """
    attrs: dict[str, str | list [str]]
    """
    Attributes of the tag.
    """

AnalysisOptions = ParamSpec('AnalysisOptions')
DocumentType: TypeVar = TypeVar('DocumentType')
TagPredicate: TypeAlias = Callable[[str | TagType], bool]
"""
Callable taking a tag name or a tag and deciding if the tag should be searched for archive.org links.
"""

AllTags: TagPredicate = lambda _: True


class AnalyseDocument(Protocol):
    """
    Takes the content of a html file and an optional tag predicate to search the specified content for archive.org links.
    """
    def __call__(self, content: bytes, tag_predicate: TagPredicate = AllTags) -> Iterable[ArchiveLink]: ...


class AnalyseURL(Protocol):
    """
    Takes an url and an optional tag predicate to search the url's content for archive.org links.
    """

    def __call__(self, url: str, tag_predicate: TagPredicate = AllTags) -> Iterable[ArchiveLink]: ...


def create_document_analyser(
    parse_document: Callable[[bytes], DocumentType],
    get_elements: Callable[[DocumentType, TagPredicate], Iterable[TagType]],
) -> AnalyseDocument:
    def analyse(content: bytes, tag_predicate: TagPredicate = lambda _: True) -> Iterable[ArchiveLink]:
        for tag in get_elements(parse_document(content), tag_predicate):
            for attribute in tag.attrs.values():
                for link in find_links(attribute):
                    yield link

    return analyse


def create_url_analyser(
        parse_document: Callable[[bytes], DocumentType],
        get_elements: Callable[[DocumentType, TagPredicate], Iterable[TagType]],
) -> AnalyseURL:

    def analyse(url: str, tag_predicate: TagPredicate = lambda _: True) -> Iterable[ArchiveLink]:
        with get(url) as response:
            for tag in get_elements(parse_document(response.content), tag_predicate):
                for attribute in tag.attrs.values():
                    for link in find_links(attribute):
                        yield link

    return analyse
